#include "uart.h"
#include "ring_buffer.h"
#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/delay.h>

struct ring_buffer rx_buffer;
struct ring_buffer tx_buffer;

// Interruption indiquant que le bus est disponible pour le prochain octet
ISR(USART_UDRE_vect) {
  if (ring_buffer_available_bytes(&tx_buffer)) {
    // Si il y a un octet à transmettre dans le buffer de transmission,
    // initialise ce transfert
    UDR0 = ring_buffer_get(&tx_buffer);
  } else {
    // Autrement, l'interruption est désactivée
    // Sans cela, elle se déclencherait en boucle car nous ne démarrons pas une
    // transmission
    UCSR0B &= ~(1 << UDRIE0);
  }
}

// Interruption indiquant qu'un octet est arrivé via l'UART
ISR(USART_RX_vect) { ring_buffer_put(&rx_buffer, UDR0); }

void uart_init(uint32_t baudrate) {
  uint16_t ubbr = F_CPU / 8 / baudrate - 1;

  // Définit la vitesse de transmission
  UBRR0H = (uint8_t)(ubbr >> 8);
  UBRR0L = (uint8_t)ubbr;

  // Active le doublement de la vitesse de transmission
  UCSR0A = (1 << U2X0);
  // Active la réception et la transmission, active l'interruption de reception
  UCSR0B = (1 << RXEN0) | (1 << TXEN0) | (1 << RXCIE0);
  // 1 stop bit, 8 data bits
  UCSR0C = (1 << USBS0) | (3 << UCSZ00);
}

void uart_send_byte(uint8_t data) {
  // Place l'octet dans le buffer circulaire d'émission
  ring_buffer_put(&tx_buffer, data);

  // On s'assure que l'interruption indiquant que le bus est disponible pour la
  // transmission est activée
  UCSR0B |= (1 << UDRIE0);
}

void uart_send_string(const char *str) {
  while (*str) {
    uart_send_byte(*str);
    str++;
  }
}

uint8_t uart_available_bytes(struct ring_buffer *rb) {
  return ring_buffer_available_bytes(&rx_buffer);
}

uint8_t uart_read_byte() { return ring_buffer_get(&rx_buffer); }
